/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ns3/debug.h"
#include "ns3/fw-simple.h"
#include "ns3/internet-node.h"
#include "ns3/serial-channel.h"
#include "ns3/datagram-socket.h"
#include "ns3/node-list.h"

#include "ns3/arp-ipv4-interface.h"
#include "ns3/ipv4.h"
#include "ns3/ipv4-address.h"
#include "ns3/datagram-socket.h"
#include "ns3/ipv4-route.h"
#include "ns3/serial-topology.h"

using namespace ns3;

class Simulation : public SimpleSimulationFramework
{
public:
  InternetNode   *CreateNode(std::string name);

  DatagramSocket *CreateSocketApp (InternetNode *node);

  InternetNode   *m_n0, *m_n1, *m_n2, *m_n3;
  SerialChannel  *m_ch1, *m_ch2, *m_ch3;
  DatagramSocket *m_src0, *m_src3, *m_sink1, *m_sink3;;

  void InitTopology (void);
  void PreSimulation (void);
  void Simulate (void);
};

  InternetNode *
Simulation::CreateNode(std::string name)
{
  NS_DEBUG_UNCOND("Simulation::CreateNode");

  InternetNode *node = new InternetNode();
  m_objectContainer.Acquire (node);
  node->SetName(name);
  NodeList::Add (node);
  return node;
}

  DatagramSocket *
Simulation::CreateSocketApp (InternetNode *node)
{
  NS_DEBUG_UNCOND("Simulation::CreateSocketApp");

  DatagramSocket *socket = new DatagramSocket (node);
  m_objectContainer.Acquire (socket);
  return socket;
}

  void
Simulation::InitTopology (void)
{
  NS_DEBUG_UNCOND("Simulation::InitTopology");

  m_n0 = CreateNode("Node 0");
  m_n1 = CreateNode("Node 1");
  m_n2 = CreateNode("Node 2");
  m_n3 = CreateNode("Node 3");

  m_ch1 = SerialTopology::AddSerialLink (m_n0, "10.1.1.1", m_n2, "10.1.1.2", 
    5000000, MilliSeconds(2));

  m_ch2 = SerialTopology::AddSerialLink (m_n1, "10.1.2.1", m_n2, "10.1.2.2",
    5000000, MilliSeconds(2));

  m_ch3 = SerialTopology::AddSerialLink (m_n2, "10.1.3.1", m_n3, "10.1.3.2",
    1500000, MilliSeconds(10));
  
  m_src0 = CreateSocketApp(m_n0);
  m_src0->SetDefaultDestination ("10.1.3.2", 80);
  m_n0->GetIpv4()->SetDefaultRoute ("10.1.1.2", 1);

  m_src3 = CreateSocketApp(m_n3);
  m_src3->SetDefaultDestination ("10.1.2.1", 80);
  m_n3->GetIpv4()->SetDefaultRoute ("10.1.3.1", 1);

  m_sink1 = CreateSocketApp(m_n3);
  m_sink1->Bind (80);

  m_sink3 = CreateSocketApp(m_n3);
  m_sink3->Bind (80);
}

  static void
DatagramSocketPrinter (
  DatagramSocket *socket, 
  uint32_t size, 
  Ipv4Address from, 
  uint16_t fromPort)
{
  std::cout << "Node: " << socket->GetNode()->GetId () 
            << " at=" << Simulator::Now ().GetSeconds () << "s,"
            << " rx bytes=" << size << std::endl;
}

  static void
GenerateTraffic (DatagramSocket *socket, uint32_t size)
{
  std::cout << "Node: " << socket->GetNode()->GetId () 
            << " at=" << Simulator::Now ().GetSeconds () << "s,"
            << " tx bytes=" << size << std::endl;
  socket->SendDummy (size);
}


  void
Simulation::PreSimulation (void)
{
  NS_DEBUG_UNCOND("Simulation::PreSimulation");

  SimpleSimulationFramework::PreSimulation ();

  m_sink3->SetDummyRxCallback (MakeCallback (&DatagramSocketPrinter));
  m_sink1->SetDummyRxCallback (MakeCallback (&DatagramSocketPrinter));

  for (double t = 0.5; t < 3.0; t += 0.5) {
    Simulator::Schedule (Seconds(t), &GenerateTraffic, m_src0, 
      static_cast<uint32_t>(100));

    Simulator::Schedule (Seconds(t), &GenerateTraffic, m_src3,
      static_cast<uint32_t>(100));
  }
}

  void
Simulation::Simulate (void)
{
  NS_DEBUG_UNCOND("Simulation::Simulate");

  Simulator::StopAt (Seconds(3.0));
  Simulator::Run ();
}

  int 
main (int argc, int argv)
{
  NS_DEBUG_UNCOND("Simple Framework-Based Simulation");

  Simulation simulation;
  simulation.Run ();

  return 0;
}
